/*
    I2C bus implementation and handling functions.
    Copyright (C) 2010 2012 2016 2018  Alessandro Accattini

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 *****************************************************************************
 *
 *
 * $Header$
 *
 *
 *
 * REVISION HISTORY:
 * $Log$
 */


/***************************************************************************/
/* include vari */

#define __I2C_C
#include "i2c.h"				// functions prototypes

#include "io.h"					// hardware control
#include "trace.h"				// function call trace
#include <stdio.h>

/***************************************************************************/
/* define interni alla libreria */

#define CLOCK_DELAY	10			// regola la velocita'

  // assegnazione dei pin
#define SDA		0
#define SCL		1

/***************************************************************************/
/* macro interne alla libreria */

#define Delay(T)

#define SetScl(v)	IO_BitWrite(SCL,v)
#define GetSda()	IO_BitRead(SDA)
#define SetSda(v)	IO_BitWrite(SDA,v)
#define SetSdaIn()	IO_DirMaskBitWrite(SDA,INPUT)
#define SetSdaOut()	IO_DirMaskBitWrite(SDA,OUTPUT)

/***************************************************************************/
/* strutture contenute all'interno della libreria */

/***************************************************************************/
/* costanti interne alla libreria */

/***************************************************************************/
/* variabili interne alla libreria */

static unsigned long int	I2cPageSize;	// dim. pagina
static unsigned int		I2cAddrBytes;	// num. byte di indirizzo
static unsigned int		I2cAddrBits;	// num. byte di indirizzo
static unsigned char		I2cDevId;	// device ID nel primo byte
static unsigned char		I2cHighMask;
static unsigned char		CtrlByte;	// byte di controllo

/***************************************************************************/
/* funzioni interne alla libreria */

static void SendStart(void);
  /* Invia la sequenza di START sul bus. */

static void SendStop(void);
  /* Invia la sequenza di STOP sul bus. */

static int SendByte(unsigned char Data);
  /* Invia un byte al device, ritorna il bit di ACK ricevuto dal device. */

static unsigned char RecvByte(unsigned char Ack);
  /* Riceve un byte dal device. */

static int OpenWrite(unsigned long int Addr);
  /* Inizia un'operazione di scrittura sul device, ritorna 0 in caso di
     successo, 1 se mancano gli ACK da parte del device. */

/***************************************************************************/
/* funzioni */

static void SendStart(void)
{
  /* Invia la sequenza di START sul bus. */
  
  
  
  FUNCTION_ENTER();
  MESSAGE(BUS_VERB_LEVEL,"I2C <- Start");

  SetSda(1);
  SetScl(1);					// alza SCL
  Delay(CLOCK_DELAY*2);  
  SetSda(0);					// abbassa SDA
  Delay(CLOCK_DELAY*2);  
  SetScl(0);					// abbassa SCL
  
  FUNCTION_LEAVE();
  return;					// termina
}



static void SendStop(void)
{
  /* Invia la sequenza di STOP sul bus. */
  
  
  
  FUNCTION_ENTER();
  
  SetScl(0);
  SetSda(0);
  SetScl(1);					// alza SCL
  Delay(CLOCK_DELAY*2);  
  SetSda(1);					// alza SDA
  Delay(CLOCK_DELAY*2);  
  
  MESSAGE(BUS_VERB_LEVEL,"I2C <- Stop");
  FUNCTION_LEAVE();
  return;
}



static int SendByte(unsigned char Data)
{
  /* Invia un byte al device, ritorna il bit di ACK ricevuto dal device. */
  
  
  
  int	i;
  int	Ack=1;			        // ACK bit
  
  
  
  FUNCTION_LEAVE();
  MESSAGE(BUS_VERB_LEVEL,"I2C <- 0x%02x  ",(unsigned int)Data);

  for(i=0;i<8;++i)
  {
    SetSda(Data&0x80);			// imposta la linea SDA
    Delay(CLOCK_DELAY);  
    SetScl(1);				// alza SCL per validare il dato sul bus
    Delay(CLOCK_DELAY);  
    SetScl(0);				// abbassa SCL
    Data<<=1;				// shifta data
  }
  SetSda(0);				// abbassa SDA
  SetSdaIn();				// imposta la linea in ingresso
  Delay(CLOCK_DELAY);  
  SetScl(1);				// alza SCL
  Delay(CLOCK_DELAY);  
  Ack=GetSda();
  SetScl(0);				// abbassa SCL
  SetSdaOut();				// imposta la linea in uscita

  MESSAGE(BUS_VERB_LEVEL,"Ack=%i\n",(unsigned int)Ack);
  FUNCTION_LEAVE();
  return(Ack);				// termina
}



static unsigned char RecvByte(unsigned char Ack)
{
  /* Riceve un byte dal device. */
  
  
  
  int		i;
  unsigned char	Byte=0x00;		// received data byte
  
  

  FUNCTION_ENTER();
  
  SetSdaIn();				// SDA in ingresso
  for(i=0;i<8;++i)
  {
    Delay(CLOCK_DELAY);  
    SetScl(1);				// alza SCL
    Delay(CLOCK_DELAY);  
    Byte<<=1;				// fa posto al nuovo bit
    Byte|=GetSda();			// aggiunge il nuovo bit
    SetScl(0);				// abbassa SCL
  }
  Delay(CLOCK_DELAY);  
  SetSdaOut();				// imposta SDA in uscita
  SetSda(Ack);				// invia ACK
  Delay(CLOCK_DELAY);  
  SetScl(1);				// alza SCL
  Delay(CLOCK_DELAY);  
  SetScl(0);				// abbassa SCL
  SetSda(0);				// abbassa SDA
  
  MESSAGE(BUS_VERB_LEVEL,"I2C -> 0x%02x  Ack=%i\n",
  	  (unsigned int)Byte,(unsigned int)Ack);
  FUNCTION_LEAVE();
  return(Byte);				// termina
}



static int OpenWrite(unsigned long int Addr)
{
  /* Inizia un'operazione di scrittura sul device, ritorna 0 in caso di
     successo, 1 se mancano gli ACK da parte del device. */
  
  
  
  unsigned int	AddrBytes=I2cAddrBytes;	// num. byte indirizzo
  unsigned char	AddrMask=I2cHighMask;	// maschera byte di indirizzo
  unsigned char	Byte;			// byte inviato al device



  // compone il control byte
  CtrlByte=(unsigned char)(Addr>>(I2cAddrBits-4));	// 3 bit piu' alti
  CtrlByte&=0x0e;
  CtrlByte|=I2cDevId;

  // invia il control byte
  SendStart();				// invia start
  if(SendByte(CtrlByte))		// se manca ACK
  {
    SendStop();				// ferma tutto
    return 1;				// termina con fallimento
  }

  // invia tutti i byte di indirizzo
  while(AddrBytes)
  {
    --AddrBytes;			// scala il contatore
    Byte=(unsigned char)(Addr>>(AddrBytes<<3));
    Byte&=AddrMask;			// maschera il primo byte
    AddrMask=0xff;			// nn maschera piu'
    if(SendByte(Byte))			// se manca ACK
    {
      SendStop();			// ferma tutto
      return 1;				// termina con fallimento
    }
  }
  return 0;				// termina con successo
}



int I2C_InitBus(unsigned long int PageSize,unsigned int AddressBits,
		unsigned char DevMask)
{
  /* Imposta la direzione e lo stato iniziale delle linee del bus,
     definisce il numero di bit di indirizzo e l'ID per l'indirizzamento
     del device sul bus. */

  
  
  int		AddrBits;		// num. bit di indirizzo
  int		HighAddrBits;		// num. bit nel byte addr. high



  // salva i parametri del bus
  I2cPageSize=PageSize;
  I2cAddrBits=AddressBits;
  I2cDevId=DevMask;
  
  // determina il numero di byte che compongono l'indirizzo
  AddrBits=I2cAddrBits-3;		// toglie i 3 pin di ID hardware
  I2cAddrBytes=AddrBits/8;		// numero byte interi
  HighAddrBits=AddrBits%8;		// il resto dei bit va qui
  if(HighAddrBits)			// se c'e' il resto
    ++I2cAddrBytes;			// aggiungi un byte
  else
    HighAddrBits=8;			// riempie il primo byte
  I2cHighMask=~(0xff<<HighAddrBits);	// maschera bit nel HIGH byte
  MESSAGE(BUS_VERB_LEVEL,
  	  "I2C: AddrBits=%i  AddrBytes=%i  HighMask=0x%02x  PageSize=%li\n",
	  I2cAddrBits,I2cAddrBytes,I2cHighMask,I2cPageSize);

  // imposta lo stato iniziale del bus
  IO_SetBaudRate(1000000);			// bus speed
  IO_DirMaskWrite((1<<SDA)|(1<<SCL));		// all outputs
  SendStop();					// bus a riposo
  return 0;					// termina con successo
}



int I2C_WriteByte(unsigned long int Address,unsigned char Data)
{
  /* Scrive un byte all'indirizzo indicato, ritorna 0 in caso di successo. */
  
  
  
  if(OpenWrite(Address))		// se il device non risponde
  {
    SendStop();				// ferma tutto
    return 1;				// termina con fallimento
  }

  // invia il data byte
  if(SendByte(Data))			// se manca ACK
  {
    SendStop();				// ferma tutto
    return 1;				// termina con fallimento
  }
  SendStop();				// invia stop
  return 0;				// termina con successo
}



int I2C_ReadByte(unsigned long int Address,unsigned char *Data)
{
  /* Legge un byte dall'indirizzo indicato, ritorna 0 in caso di successo. */



  // esegue un dummy write
  if(OpenWrite(Address))		// se il device non risponde
  {
    SendStop();				// ferma tutto
    return 1;				// termina con fallimento
  }
  
  // inizia una nuova transazione
  SendStart();				// invia start
  if(SendByte(CtrlByte|0x01))		// se manca ACK
  {
    SendStop();				// ferma tutto
    return 1;				// termina con fallimento
  }

  // legge 1 byte dal device
  *Data=RecvByte(1);			// riceve un solo byte
  SendStop();				// invia STOP
  return 0;				// termina con sucesso
}



int I2C_WritePage(unsigned long int Address,const unsigned char *Data)
{
  /* Scrive una pagina all'indirizzo indicato, ritorna 0 in caso di successo. */
  
  

  unsigned long int	i=I2cPageSize;
 


  if(OpenWrite(Address))		// se il device non risponde
  {
    SendStop();				// ferma tutto
    return 1;				// termina con fallimento
  }

  // invia la pagina
  while(i)
  {
    if(SendByte(*Data))			// se manca ACK
    {
      SendStop();			// ferma tutto
      return 1;				// termina con fallimento
    }
    Data++;				// next byte
    i--;				// scala il contatore
  }
  SendStop();				// invia stop
  return 0;				// termina con successo
}



int I2C_ReadPage(unsigned long int Address,unsigned char *Data)
{
  /* Legge una pagina dall'indirizzo indicato, ritorna 0 in caso di successo. */
  
  
  
  unsigned long int	i=I2cPageSize;



  // esegue un dummy write
  if(OpenWrite(Address))		// se il device non risponde
  {
    SendStop();				// ferma tutto
    return 1;				// termina con fallimento
  }
  
  // inizia una nuova transazione
  SendStart();				// invia start
  if(SendByte(CtrlByte|0x01))		// se manca ACK
  {
    SendStop();				// ferma tutto
    return 1;				// termina con fallimento
  }

  // legge una pagina dal device
  while(i)
  {
    i--;				// scala il contatore
    *Data=RecvByte(!i);			// riceve un byte
    Data++;				// next array entry
  }
  SendStop();				// invia STOP
  return 0;				// termina con sucesso
}

/* $Source$ */

