/*
    I/O level handling functions.
    Copyright (C) 2010 2013 2016 2018  Alessandro Accattini

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 *****************************************************************************
 *
 *
 * $Header$
 *
 *
 *
 * REVISION HISTORY:
 * $Log$
 */


/***************************************************************************/
/* include vari */

#define __IO_C
#include "io.h"				// functions prototypes

#include <stdio.h>			// fprintf
#include <ftdi.h>			// FTDI lib
#include "trace.h"			// function call trace

/***************************************************************************/
/* define interni alla libreria */

/***************************************************************************/
/* strutture contenute all'interno della libreria */

/*
typedef union
{
  unsigned long int	All;		// all bits at once
  unsigned char		Byte[4];	// one byte at a time
}IOdata_t;
*/

/***************************************************************************/
/* costanti interne alla libreria */

/***************************************************************************/
/* variabili interne alla libreria */

static struct ftdi_context	Hard;		// local hardware context
static BusData_t		Output=0x00;	// bit in uscita
static BusData_t		DirMask=0xff;	// maschera dir. I/O
static BusData_t		Input;		// ultmo ingresso letto
static int			Interfaces;	// totale interfacce presenti

/***************************************************************************/
/* funzioni interne alla libreria */

static int ChangeLatencyTimer(struct ftdi_context *Context,
				unsigned char NewLatency);
  /* Modifica il tempo di latenza del device specificato, ritorna 0 in
     caso di successo. */

/***************************************************************************/
/* funzioni */

static int ChangeLatencyTimer(struct ftdi_context *Context,
				unsigned char NewLatency)
{
  /* Modifica il tempo di latenza del device specificato, ritorna 0 in
     caso di successo. */



  unsigned char		Latency;		// tatenza originale



  FUNCTION_ENTER();
  if(ftdi_get_latency_timer(Context,&Latency))	// se fallisce
  {
    fprintf(stderr,"%s: ftdi_get_latency_timer failed: %s.\n",
    			__FUNCTION__,ftdi_get_error_string(Context));
    return 1;				// termina con fallimento
  }
  if(ftdi_set_latency_timer(Context,NewLatency))	// se fallisce
  {
    fprintf(stderr,"%s: ftdi_set_latency_timer failed: %s.\n",
    			__FUNCTION__,ftdi_get_error_string(Context));
    return 1;				// termina con fallimento
  }
  printf("Original latency timer: %u changed to: %u\n",
  		(unsigned int)Latency,
		(unsigned int)NewLatency);
  FUNCTION_LEAVE();
  return 0;				// termina con successo
}



int IO_Open(int Vendor,int Product,int HardIfaces)
{
  /* Inizializza il controllo diretto delle linee di I/O e imposta lo stato
     iniziale, ritorna 0 in caso di successo. */
  /**< \brief Opens and initilizes the hardware level.
       \param Vendor
       \param Product
       \param HardIfaces
       \return returns 0 on success.

       This function initializes the hardware level, must be called before
       any other function.
   */
  
  
  
  struct ftdi_device_list	*DevList;	// devices trovati
  struct ftdi_device_list	*CurDev;	// device attuale
  int				Success;
  int				DevNum;
  int				i=0;
  char				Manuf[128];	// device manufacturer
  char				Desc[128];	// device description
  char				Sn[128];	// device serial number



  FUNCTION_ENTER();
  
  // inizializza il bus
  if(ftdi_init(&Hard) < 0)
  {
    fprintf(stderr,"%s: ftdi_init failed.\n",__FUNCTION__);
    return 1;				// termina con fallimento
  }

  // esegue una ricerca di tutti i devices collegati
  if((DevNum=ftdi_usb_find_all(&Hard,&DevList,Vendor,Product))<0)
  {
    fprintf(stderr,"%s: ftdi_usb_find_all failed: %s.\n",
    			__FUNCTION__,ftdi_get_error_string(&Hard));
    return 1;				// termina con fallimento
  }

  // stampa la lista di devices trovati
  printf("Found %i FTDI devices:\n",DevNum);
  if(!DevNum)				// se non ci sono devices
    return 1;				// termina con fallimento
  CurDev=DevList;			// punta al primo device della lista
  do
  {
    printf("Device: %i, ",i++);
    if(ftdi_usb_get_strings(&Hard,CurDev->dev,Manuf,128,Desc,128,Sn,128)<0)
    {
      fprintf(stderr,"%s: ftdi_usb_get_strings failed: %s.\n",
			__FUNCTION__,ftdi_get_error_string(&Hard));
      return 1;				// termina con fallimento
    }
    printf("manufacturer: %s, description: %s, serial: %s\n",Manuf,Desc,Sn);
    CurDev=CurDev->next;		// passa al prossimo device
  }while(CurDev!=NULL);			// fino a fine lista
  ftdi_list_free(&DevList);		// dealloca la lista

  // determina il numero di interfacce hardware a 8 bit
  if(!HardIfaces)
  {
    printf("%s: type=%i\n",__FUNCTION__,(int)Hard.type);
    Interfaces=1;			// una c'e' di sicuro
    if((Hard.type == TYPE_2232C) || (Hard.type == TYPE_2232H))
      Interfaces=2;			// 2 interfacce a 8 bit ciascuna
    if(Hard.type == TYPE_4232H)
      Interfaces=4;			// 4 interfacce a 8 bit ciascuna
  }
  else
  {
    Interfaces=HardIfaces;		// forza il numero di interfacce
  }
  printf("Found hardware interfaces: %i\n",Interfaces);

  // apre le interfacce hardware trovate
  if(Interfaces >= 1)
  {
    ftdi_set_interface(&Hard,INTERFACE_A);	// prima interfaccia hardware
    Success=ftdi_usb_open(&Hard,Vendor,Product);
    if((Success<0)&&(Success!=-5))
    {
      fprintf(stderr,"%s: unable to open ftdi device: %s\n",
			  __FUNCTION__,ftdi_get_error_string(&Hard));
      return 1;				// temina con fallimento	
    }
    ChangeLatencyTimer(&Hard,2);	// abbassa la latenza
  }



/*  
  if(Interfaces >= 2)
  {
    if(ftdi_init(&Hard[1]) < 0)
    {
      fprintf(stderr,"%s: ftdi_init failed.\n",__FUNCTION__);
      return 1;				// termina con fallimento
    }
    ftdi_set_interface(&Hard[1],INTERFACE_B);	// seconda interfaccia hardware
    Success=ftdi_usb_open(&Hard[1],Vendor,Product);
    if((Success<0)&&(Success!=-5))
    {
      fprintf(stderr,"%s: unable to open ftdi device: %s\n",
			  __FUNCTION__,ftdi_get_error_string(&Hard[1]));
      return 1;				// temina con fallimento	
    }
    ChangeLatencyTimer(&Hard[1],2);	// abbassa la latenza
  }
  if(Interfaces >= 4)
  {
    if(ftdi_init(&Hard[2]) < 0)
    {
      fprintf(stderr,"%s: ftdi_init failed.\n",__FUNCTION__);
      return 1;				// termina con fallimento
    }
    ftdi_set_interface(&Hard[2],INTERFACE_C);	// terza interfaccia hardware
    Success=ftdi_usb_open(&Hard[2],Vendor,Product);
    if((Success<0)&&(Success!=-5))
    {
      fprintf(stderr,"%s: unable to open ftdi device: %s\n",
			  __FUNCTION__,ftdi_get_error_string(&Hard[2]));
      return 1;				// temina con fallimento	
    }
    ChangeLatencyTimer(&Hard[2],2);	// abbassa la latenza

    if(ftdi_init(&Hard[3]) < 0)
    {
      fprintf(stderr,"%s: ftdi_init failed.\n",__FUNCTION__);
      return 1;				// termina con fallimento
    }
    ftdi_set_interface(&Hard[3],INTERFACE_D);	// quarta interfaccia hardware
    Success=ftdi_usb_open(&Hard[3],Vendor,Product);
    if((Success<0)&&(Success!=-5))
    {
      fprintf(stderr,"%s: unable to open ftdi device: %s\n",
			  __FUNCTION__,ftdi_get_error_string(&Hard[3]));
      return 1;				// temina con fallimento	
    }
    ChangeLatencyTimer(&Hard[3],2);	// abbassa la latenza
  }
*/

  // imposta lo stato fondamentale di tutta l'interfaccia hardware
  IO_DirMaskWrite(DirMask);		// imposta la dir iniziale dei pin
  IO_DataWrite(Output);			// imposta lo stato iniziale
  FUNCTION_LEAVE();
  return 0;				// termina con successo
}



int IO_Close(void)
{
  /* Rilascia le risorse allocate e chiude l'interfaccia con l'hardware. */
  
  
  
  FUNCTION_ENTER();

  ftdi_disable_bitbang(&Hard);
  ftdi_usb_close(&Hard);
  ftdi_deinit(&Hard);

  FUNCTION_LEAVE();
  return 0;				// termina con successo
}



int IO_SetBaudRate(int Baud)
{
  /* Imposta il baudrate di aggiornamento dell'I/O. */
  
  
  
  FUNCTION_ENTER();
  MESSAGE(IO_VERB_LEVEL,"IO.baudrate <- %i\n",Baud);
  
  ftdi_set_baudrate(&Hard,Baud);	// imposta il baudrate
  
  FUNCTION_LEAVE();
  return 0;				// termina con successo
}



BusData_t IO_DirMaskRead(void)
{
  /* Restituisce la maschera direzioni corrente. */
  
  
  
  FUNCTION_ENTER();
  MESSAGE(IO_VERB_LEVEL,"IO.dir -> 0x%02lx\n",(unsigned long int)DirMask);
  FUNCTION_LEAVE();

  return(DirMask);			// termina
}



int IO_DirMaskWrite(BusData_t Mask)
{
  /* Imposta la direzione dei pin di I/O, ritorna 0 in caso di sucesso. */
  
 

  FUNCTION_ENTER();
  MESSAGE(IO_VERB_LEVEL,"IO.dir <- 0x%02lx\n",(unsigned long int)Mask);
  
  DirMask=Mask;				// salva la dir mask completa
  if(ftdi_set_bitmode(&Hard,DirMask,BITMODE_BITBANG) < 0)
    fprintf(stderr,"%s: set pins direction (%s)\n",
		__FUNCTION__,ftdi_get_error_string(&Hard));

  FUNCTION_LEAVE();
  return 0;				// termina con successo
}



int IO_DirMaskSet(BusData_t SetMask)
{
  /* Setta i bit impostati nella maschera, ritorna 0 in caso di sucesso. */
  
  
  
  FUNCTION_ENTER();
  FUNCTION_LEAVE();

  return(IO_DirMaskWrite(DirMask | SetMask));
}



int IO_DirMaskReset(BusData_t ResetMask)
{
  /* Resetta i bit impostati nella maschera, ritorna 0 in caso di sucesso. */
  
  
  
  FUNCTION_ENTER();
  FUNCTION_LEAVE();

  return(IO_DirMaskWrite(DirMask & (~ResetMask)));
}



int IO_DirMaskToggle(BusData_t ToggleMask)
{
  /* Inverte i bit impostati nella maschera, ritorna 0 in caso di sucesso. */
  
  
  
  FUNCTION_ENTER();
  FUNCTION_LEAVE();

  return(IO_DirMaskWrite(DirMask ^ ToggleMask));
}



BusData_t IO_DirMaskBitRead(unsigned short int Bit)
{
  /* Restituisce la direzione del pin indicato. */
  
  
  
  BusData_t	DirBit;			// bit prelevato dalla maschera



  FUNCTION_ENTER();
  
  DirBit=(DirMask >> Bit) & 0x01;	// valore singolo bit
  
  MESSAGE(IO_VERB_LEVEL,"IO.dir[%hi] -> %i\n",Bit,(unsigned int)DirBit);
  FUNCTION_LEAVE();
  return(DirBit);			// termina
}



int IO_DirMaskBitWrite(unsigned short int Bit,BusData_t NewDir)
{
  /* Imposta la direzione dei pin indicato, ritorna 0 in caso di sucesso. */
  
  
  
  FUNCTION_ENTER();
  MESSAGE(IO_VERB_LEVEL,"IO.dir[%hi] <- %i\n",Bit,(unsigned int)NewDir);
  FUNCTION_LEAVE();

  return(IO_DirMaskWrite((DirMask & ~(1<<Bit)) || ((NewDir & 0x01) << Bit)));
}



int IO_DirMaskBitSet(unsigned short int Bit)
{
  /* Setta la direzione dei pin indicato, ritorna 0 in caso di sucesso. */
  
  
  
  FUNCTION_ENTER();
  MESSAGE(IO_VERB_LEVEL,"IO.dir[%hi] <- 1\n",Bit);
  FUNCTION_LEAVE();

  return(IO_DirMaskSet(1 << Bit));
}



int IO_DirMaskBitReset(unsigned short int Bit)
{
  /* Resetta la direzione dei pin indicato, ritorna 0 in caso di sucesso. */
  
  
  
  FUNCTION_ENTER();
  MESSAGE(IO_VERB_LEVEL,"IO.dir[%hi] <- 0\n",Bit);
  FUNCTION_LEAVE();

  return(IO_DirMaskReset(1 << Bit));
}



int IO_DirMaskBitToggle(unsigned short int Bit)
{
  /* Inverte la direzione dei pin indicato, ritorna 0 in caso di sucesso. */
  
  
  
  FUNCTION_ENTER();
  MESSAGE(IO_VERB_LEVEL,"IO.dir[%hi] <- t\n",Bit);
  FUNCTION_LEAVE();

  return(IO_DirMaskToggle(1 << Bit));
}



BusData_t IO_OutDataRead(void)
{
  /* Legge lo stato di tutti i pin in uscita. */
  
  
  
  FUNCTION_ENTER();
  MESSAGE(IO_VERB_LEVEL,"IO.out -> 0x%02lx\n",(unsigned long int)Output);
  FUNCTION_LEAVE();

  return(Output);			// termina
}



BusData_t IO_DataRead(void)
{
  /* Legge lo stato di tutti i pin in ingresso. */



  FUNCTION_ENTER();
  
  if(ftdi_read_pins(&Hard,&Input) < 0)		// tenta la lettura
    fprintf(stderr,"%s: read data failed (%s)\n",
		__FUNCTION__,ftdi_get_error_string(&Hard));
  
  MESSAGE(IO_VERB_LEVEL,"IO.in -> 0x%02lx\n",(unsigned long int)Input);
  FUNCTION_LEAVE();
  return(Input);				// termina
}



void IO_DataWrite(BusData_t DataMask)
{
  /* Imposta lo stato di tutti i pin in uscita. */
  
  

  FUNCTION_ENTER();
  MESSAGE(IO_VERB_LEVEL,"IO.out <- 0x%02lx\n",(unsigned long int)DataMask);
  
  Output=DataMask;				// salva lo stato delle uscite
  if(ftdi_write_data(&Hard,&Output,1) < 0)
    fprintf(stderr,"%s: write data failed (%s)\n",
		__FUNCTION__,ftdi_get_error_string(&Hard));
  
  FUNCTION_LEAVE();
  return;					// termina
}



void IO_Set(BusData_t SetMask)
{
  /* Setta i bit impostati nella maschera. */
  
  
  
  FUNCTION_ENTER();
  
  IO_DataWrite(Output | SetMask);
  
  FUNCTION_LEAVE();
  return;					// termina
}



void IO_Reset(BusData_t ResetMask)
{
  /* Resetta i bit impostati nella maschera. */
  
  
  
  FUNCTION_ENTER();
  
  IO_DataWrite(Output & ~(ResetMask));
  
  FUNCTION_LEAVE();
  return;					// termina
}



void IO_Toggle(BusData_t ToggleMask)
{
  /* Inverte i bit impostati nella maschera. */
  
  
  
  FUNCTION_ENTER();
  
  IO_DataWrite(Output ^ ToggleMask);
  
  FUNCTION_LEAVE();
  return;					// termina
}



BusData_t IO_BitRead(unsigned short int Bit)
{
  /* Legge lo stato del pin indicato. */



  BusData_t	InBit;



  FUNCTION_ENTER();
  
  InBit=(IO_DataRead() >> Bit) & 0x01;		// tiene un solo bit
  
  MESSAGE(IO_VERB_LEVEL,"IO.in[%hi] -> %i\n",Bit,(unsigned int)InBit);
  FUNCTION_LEAVE();
  return(InBit);				// termina
}



void IO_BitWrite(unsigned short int Bit,BusData_t NewValue)
{
  /* Imposta lo stato del pin indicato. */
  
  
  
  FUNCTION_ENTER();
  MESSAGE(IO_VERB_LEVEL,"IO.out[%hi] <- %i\n",Bit,(unsigned int)NewValue);
  
  IO_DataWrite((Output & ~(1<<Bit)) | ((NewValue & 0x01) << Bit));
  
  FUNCTION_LEAVE();
  return;					// termina
}



void IO_BitSet(unsigned short int Bit)
{
  /* Setta il bit indicato. */
  
  
  
  FUNCTION_ENTER();
  MESSAGE(IO_VERB_LEVEL,"IO.out[%hi] <- 1\n",Bit);
  
  IO_Set(1 << Bit);
  
  FUNCTION_LEAVE();
  return;					// termina
}



void IO_BitReset(unsigned short int Bit)
{
  /* Resetta il bit indicato. */
  
  
  
  FUNCTION_ENTER();
  MESSAGE(IO_VERB_LEVEL,"IO.out[%hi] <- 0\n",Bit);

  IO_Reset(1 << Bit);
  
  FUNCTION_LEAVE();
  return;					// termina
}



void IO_BitToggle(unsigned short int Bit)
{
  /* Inverte il bit indicato. */
  
  
  
  FUNCTION_ENTER();
  MESSAGE(IO_VERB_LEVEL,"IO.out[%hi] <- t\n",Bit);
  
  IO_Toggle(1 << Bit);
  
  FUNCTION_LEAVE();
  return;					// termina
}

/* $Source$ */

