/*
    I/O level handling functions.
    Copyright (C) 2010 2013 2015 2016 2018  Alessandro Accattini

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 *****************************************************************************
 *
 *
 * $Header$
 *
 *
 *
 * REVISION HISTORY:
 * $Log$
 */

/**  \file    io.h
     \author  Alessandro Accattini
     \date    2 Gen 2018
     \brief   Basic I/O level handling functions.

     I/O level is the lowest available level, at this level it is possible to
     handle the pins direction, moving up and down output pins and reading
     data on input pins. All the other interfaces and buses, like SPI and I2C,
     are built on top of this level.
 */


#ifndef __IO_H
#define __IO_H	1


/***************************************************************************/
/* define pubblici */

  // pin direction
#define INPUT			0		// pin in ingresso
#define OUTPUT			1		// pin in uscita


#define IO_BUS_WIDTH		8		// larghezza del bus in bit

/***************************************************************************/
/* macro pubbliche */

/***************************************************************************/
/* strutture pubbliche */

/** \brief  Bus data type.
 *
 *  Hardware bus is represented by 8 data lines; each line can be configured
 *  as input or output indipendently from the others. We need a data type to
 *  represent the full bus at once.
 */
typedef unsigned char		BusData_t;

/***************************************************************************/
/* costanti pubbliche */

/***************************************************************************/
/* variabili pubbliche */

/***************************************************************************/
/* funzioni pubbliche */

int IO_Open(int Vendor,int Product,int HardIfaces);
  /* Inizializza il controllo diretto delle linee di I/O e imposta lo stato
     iniziale, ritorna 0 in caso di successo. */
  /**< \brief  Opens and initilizes the hardware level.
       \param  Vendor
       \param  Product
       \param  HardIfaces
       \return returns 0 on success.

       This function initializes the hardware level, must be called before
       any other function.
   */

int IO_Close(void);
  /**< \brief  Close and release the hardware level.
       \return returns 0 on success.

       This function close the hardware interface and free all the resources;
       after this call it not possible to call any other function.
   */


int IO_SetBaudRate(int Baud);
  /**< \brief  Change the hardware baudrate.
       \param  Baud  the new baudrate.
       \return returns 0 on success.

       Changing baudrate can influence the bus speed.
   */


BusData_t IO_DirMaskRead(void);
  /**< \brief  Reads back the full bus direction mask.
       \return the bus direction mask.

       In this way we can read back the full bus direction mask: each bit in
       the mask defines the direction of the corresponding data line.
   */

int IO_DirMaskWrite(BusData_t DirMask);
  /**< \brief  Set the full bus direction mask.
       \param DirMask  new direction mask.
       \return returns 0 on success.

       This function overwrite the full bus direction mask.
   */

int IO_DirMaskSet(BusData_t SetMask);
  /**< \brief  Set data lines as output.
       \param SetMask  direction mask.
       \return returns 0 on success.

       This function sets only bits set to 1 in the SetMask, this means lines
       set to 1 in the SetMask are routed as outputs.
   */

int IO_DirMaskReset(BusData_t ResetMask);
  /**< \brief  Set data lines as input.
       \param ResetMask  direction mask.
       \return returns 0 on success.

       This function resets only bits set to 1 in the ResetMask, this means
       lines set to 1 in the ResetMask are routed as input.
   */

int IO_DirMaskToggle(BusData_t ToggleMask);
  /**< \brief  Toggle data lines direction.
       \param ToggleMask  direction mask.
       \return returns 0 on success.

       This function toggle the direction of lines set to 1 in the ToggleMask.
   */


BusData_t IO_DirMaskBitRead(unsigned short int Bit);
  /* Restituisce la direzione del pin indicato. */
  /**< \brief  Reads back a single data line direction.
       \param Bit  single data line number.
       \return data line direction mask.

       This function can be useful to handle the direction of a single data
       line.
   */

int IO_DirMaskBitWrite(unsigned short int Bit,BusData_t NewDir);
  /**< \brief  Writes a single data line direction.
       \param Bit  single data line number.
       \param NewDir  new data line direction.
       \return returns 0 on success.

       Overwrite the direction of a single bus data line.
       This function can be useful to handle the direction of a single data
       line.
   */

int IO_DirMaskBitSet(unsigned short int Bit);
  /**< \brief  Set a single data line direction as output.
       \param Bit  single data line number.
       \return returns 0 on success.

       This function routes a single bus data line as output.
   */

int IO_DirMaskBitReset(unsigned short int Bit);
  /**< \brief  Set a single data line direction as input.
       \param Bit  single data line number.
       \return returns 0 on success.

       This function routes a single bus data line as input.
   */

int IO_DirMaskBitToggle(unsigned short int Bit);
  /**< \brief  Toggles a single data line direction.
       \param Bit  single data line number.
       \return returns 0 on success.

       This function toggles the direction of a single bus data line.
   */


BusData_t IO_OutDataRead(void);
  /**< \brief  Read back the output data register.
       \return the output data value.

       This function reads back the content of the output data register, the
       operation does not involve input data lines, simply reads back the last
       value written by the function IO_DataWrite().
   */

BusData_t IO_DataRead(void);
  /**< \brief  Read the input data register.
       \return the input data value.

       This function reads the content of the input data register, all input
       data lines are latched and the value is returned; all output data lines
       are simply read back.
   */

void IO_DataWrite(BusData_t DataMask);
  /**< \brief  Write the output data register.
       \param DataMask  new bus data value.

       This function writes the content of the output data register, all input
       data lines are simply unaffected by this operatrion; all output data
       lines are update accordingly with DataMask value.
   */

void IO_Set(BusData_t SetMask);
  /**< \brief  set bits in the output data register.
       \param SetMask  bits mask to set.

       Only bits set to 1 in the SetMask are set to 1 in the bus.
       Input data lines are not affected by this operation.
   */

void IO_Reset(BusData_t ResetMask);
  /**< \brief  reset bits in the output data register.
       \param ResetMask  bits mask to set.

       Only bits set to 1 in the ResetMask are set to 0 in the bus.
       Input data lines are not affected by this operation.
   */

void IO_Toggle(BusData_t ToggleMask);
  /**< \brief  toggle bits in the output data register.
       \param ToggleMask  bits mask to set.

       Only bits set to 1 in the ToggleMask are toggled in the bus.
       Input data lines are not affected by this operation.
   */


BusData_t IO_BitRead(unsigned short int Bit);
  /**< \brief  reads a single data line level.
       \param Bit  single data line number.
       \return  data line level.

       This function is used to read the level of a single input data line.
       Reading on an output data line returns the output value.
   */

void IO_BitWrite(unsigned short int Bit,BusData_t NewValue);
  /**< \brief  writes a single data line level.
       \param Bit  single data line number.
       \param NewValue  new data line level.

       This function is used to write the level of a single output data line.
       Writing to an input data line do nothing.
   */

void IO_BitSet(unsigned short int Bit);
  /**< \brief  set a single data line level to high.
       \param Bit  single data line number.

       This function is used to set to 1 a single output data line.
       Trying to set an input data line do nothing.
   */

void IO_BitReset(unsigned short int Bit);
  /**< \brief  reset a single data line level to low.
       \param Bit  single data line number.

       This function is used to set to 0 a single output data line.
       Trying to reset an input data line do nothing.
   */

void IO_BitToggle(unsigned short int Bit);
  /**< \brief  toggle a single data line level.
       \param Bit  single data line number.

       This function is used to toggle a single output data line level.
       Trying to toggle an input data line do nothing.
   */

/***************************************************************************/
#endif

/* $Source$ */

