/*
    PlayStation memory card driver.
    Copyright (C) 2008 2011 2012 2016  Alessandro Accattini

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 *****************************************************************************
 *
 *
 * $Header$
 *
 *
 *
 * REVISION HISTORY:
 * $Log$
 */


/***************************************************************************/
/* include vari */

#define __PSCARD_C
#include "pscard.h"				// functions prototypes

#include "io.h"					// I/O level
#include "spi.h"				// SPI channel
#include <unistd.h>				// usleep

/***************************************************************************/
/* define interni alla libreria */

/***************************************************************************/
/* macro interne alla libreria */

#define SetSEL(v)	IO_BitWrite(3,v)
#define GetACK()	IO_BitRead(7)

/***************************************************************************/
/* strutture contenute all'interno della libreria */

/***************************************************************************/
/* costanti interne alla libreria */

//static const unsigned char	BusMap[8]={0,1,2,4,5,6,7,3};

/***************************************************************************/
/* variabili interne alla libreria */

/***************************************************************************/
/* funzioni interne alla libreria */

static int ByteExchange(unsigned char Sent,unsigned char *Recv);
  /* Scambia un byte con la card e aspetta l'ack, nel caso in cui mancasse
     l'ack alza la linea SEL e ritorna 1, ritorna 0 normalmente. */

static unsigned char WaitAck(void);
  /* Legge il bit ACK, ritorna il valore di ACK. */

/***************************************************************************/
/* funzioni */

static int ByteExchange(unsigned char Sent,unsigned char *Recv)
{
  /* Scambia un byte con la card e aspetta l'ack, nel caso in cui mancasse
     l'ack alza la linea SEL e ritorna 1, ritorna 0 normalmente. */



  unsigned char	From;				// byte ricevuto dalla card



  From=SPI_DataExchange(Sent);			// invia/riceve un byte
  if(Recv!=NULL)				// se e' atteso
    *Recv=From;					// restituisce il byte ricev.
  if(WaitAck())					// se manca ACK
  {
    SetSEL(1);					// alza la linea SEL
    return 1;					// termina con fallimento
  }
  return 0;					// termina con sucesso
}



static unsigned char WaitAck(void)
{
  /* Legge il bit ACK, ritorna il valore di ACK. */



  usleep(1000);					// aspetta 1ms
  return(GetACK());				// termina
}



int PSCARD_InitBus(void *Chipset,int argc,char **argv)
{
  /* Imposta lo stato iniziale del bus, ritorna 0 in caso di successo,
     1 altrimenti. */
  
  
  
  // inizializza il bus
//  SPI_PinRemap(BusMap);				// rimappa i pin di I/O
  if(SPI_InitBus(SPI_CPOL_HI|SPI_SECOND_EDGE|SPI_LSB_FIRST,8))
    return 1;					// termina con fallimento

  // imposta la direzione dei pin non SPI
  IO_DirMaskBitSet(3);				// bit 3 in uscita
  IO_DirMaskBitReset(7);			// bit 7 in ingresso
  SetSEL(1);					// alza SEL
  return 0;					// termina con successo
}



int PSCARD_BlockWrite(unsigned long int Address,const unsigned char Data[],int Len)
{
  /* Scrive un blocco dati all'indirizzo indicato. */



  int		i;
  unsigned char	Check1,Check2;			// byte di controllo
  unsigned char	Checksum;			// checksum byte



  // convrte l'indirizo di memoria in indirizzo di pagina
  Address>>=7;					// 128 bytes/page

  // inizia la scrittura di una pagina
  SetSEL(0);					// abbassa SEL
  if(ByteExchange(0x81,NULL))			// se mnca ACK
    return 1;					// termina con fallimento
  if(ByteExchange(0x57,NULL))			// se mnca ACK
    return 2;					// termina con fallimento
  if(ByteExchange(0x00,&Check1))		// se mnca ACK
    return 3;					// termina con fallimento
  if(ByteExchange(0x00,&Check2))		// se mnca ACK
    return 4;					// termina con fallimento
  if((Check1!=0x5a)||(Check2!=0x5d))
  {
    SetSEL(1);					// lza la linea SEL
    return 5;					// termina con fallimento
  }
  
  // seleziona la pagina scritta
  if(ByteExchange((Address>>8)&0xff,NULL))	// se mnca ACK
    return 6;					// termina con fallimento
  if(ByteExchange(Address&0xff,NULL))		// se mnca ACK
    return 7;					// termina con fallimento

  // invia la pagina alla memory card
  Checksum=(Address|(Address>>8));		// inizializza il cheksum
  for(i=0;i<128;i++)
  {
    if(ByteExchange(Data[i],NULL))		// se mnca ACK
      return 13;				// termina con fallimento
    Checksum^=Data[i];				// aggiorna la checksum
  }

  // chiude la transazione
  if(ByteExchange(Checksum,NULL))		// se mnca ACK
    return 14;					// termina con fallimento
  if(ByteExchange(0x00,&Check1))		// se mnca ACK
    return 15;					// termina con fallimento
  if(ByteExchange(0x00,&Check2))		// se mnca ACK
    return 16;					// termina con fallimento
  if((Check1!=0x5c)||(Check2!=0x5d))
  {
    SetSEL(1);					// lza la linea SEL
    return 17;					// termina con fallimento
  }
  if(SPI_DataExchange(0x00)!=0x47)		// end flag
  {
    SetSEL(1);					// lza la linea SEL
    return 18;					// termina con fallimento
  }
  SetSEL(1);					// alza SEL
  return 0;					// termina con successo
}



int PSCARD_BlockRead(unsigned long int Address,unsigned char *Data,int Len)
{
  /* Legge un blocco dati dall'indirizzo indicato. */
  
  
  
  int		i;
  unsigned char	Check1,Check2;			// byte di controllo
  unsigned char	Checksum;			// checksum byte



  // convrte l'indirizo di memoria in indirizzo di pagina
  Address>>=7;					// 128 bytes/page

  // inizia la lettura di una pagina
  SetSEL(0);					// abbassa SEL
  if(ByteExchange(0x81,NULL))			// se mnca ACK
    return 1;					// termina con fallimento
  if(ByteExchange(0x52,NULL))			// se mnca ACK
    return 2;					// termina con fallimento
  if(ByteExchange(0x00,&Check1))		// se mnca ACK
    return 3;					// termina con fallimento
  if(ByteExchange(0x00,&Check2))		// se mnca ACK
    return 4;					// termina con fallimento
  if((Check1!=0x5a)||(Check2!=0x5d))
  {
    SetSEL(1);					// lza la linea SEL
    return 5;					// termina con fallimento
  }
  
  // seleziona la pagina letta
  if(ByteExchange((Address>>8)&0xff,NULL))	// se mnca ACK
    return 6;					// termina con fallimento
  if(ByteExchange(Address&0xff,NULL))		// se mnca ACK
    return 7;					// termina con fallimento
  
  // command ack
  if(ByteExchange(0x00,&Check1))		// se mnca ACK
    return 8;					// termina con fallimento
  if(Check1!=0x5c)
  {
    SetSEL(1);					// lza la linea SEL
    return 9;					// termina con fallimento
  }
  if(ByteExchange(0x00,&Check1))		// se mnca ACK
    return 8;					// termina con fallimento
  if(Check1!=0x5d)
  {
    SetSEL(1);					// lza la linea SEL
    return 10;					// termina con fallimento
  }
  if(ByteExchange(0x00,NULL))			// se mnca ACK
    return 11;					// termina con fallimento
  if(ByteExchange(0x00,NULL))			// se mnca ACK
    return 12;					// termina con fallimento
  
  // riceve la pagina dalla memory card
  Checksum=(Address|(Address>>8));		// inizializza il cheksum
  for(i=0;i<128;i++)
  {
    if(ByteExchange(0x00,&Data[i]))		// se mnca ACK
      return 13;				// termina con fallimento
    Checksum^=Data[i];				// aggiorna la checksum
  }

  // chiude la transazione
  if(ByteExchange(0x00,NULL))			// se mnca ACK
    return 14;					// termina con fallimento
  if(SPI_DataExchange(0x00)!=0x47)		// end flag
  {
    SetSEL(1);					// lza la linea SEL
    return 15;					// termina con fallimento
  }
  SetSEL(1);					// alza SEL
  return 0;					// termina con successo
}

/* $Source$ */

